/* $NetBSD$ */

/*-
 * Copyright (c) 2004 Ben Harris
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <dev/pckbport/pckbportvar.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <arch/arm/iomd/iomdreg.h>
#include <arch/arm/iomd/iomdvar.h>
#include <arch/arm/iomd/iomdkbcvar.h>

/* XXX belongs in iomdreg.h */
#define IOMDKBC_TXE	0x80
#define IOMDKBC_TXB	0x40
#define IOMDKBC_RXF	0x20
#define IOMDKBC_RXB	0x10
#define IOMDKBC_ENA	0x08
#define IOMDKBC_RXP	0x04
#define IOMDKBC_DATA	0x02
#define IOMDKBC_CLK	0x01

#define IOMDKBC_DR	0x00
#define IOMDKBC_CR	0x01

#define KBC_DEVCMD_ACK 0xfa
#define KBC_DEVCMD_RESEND 0xfe

struct iomdkbc_softc {
	struct device	sc_dev;

	struct pckbport_tag	sc_pt;

	int			sc_haveport[PCKBPORT_NSLOTS];
	bus_space_tag_t		sc_iot;
	bus_space_handle_t	sc_ioh[PCKBPORT_NSLOTS];

	void	*sc_rxih[PCKBPORT_NSLOTS];
	int	sc_rxirq[PCKBPORT_NSLOTS];
};

static int iomdkbc_match(struct device *, struct cfdata *, void *);
static void iomdkbc_attach(struct device *, struct device *, void *);

static int iomdkbc_xt_translation(struct device *, pckbport_slot_t, int);
static int iomdkbc_send_devcmd(struct device *, pckbport_slot_t, u_char);
static int iomdkbc_poll_data1(struct device *, pckbport_slot_t);
static void iomdkbc_slot_enable(struct device *, pckbport_slot_t, int);
static void iomdkbc_intr_establish(struct device *, pckbport_slot_t);

static int iomdkbc_intr(void *);

CFATTACH_DECL(iomdkbc, sizeof(struct iomdkbc_softc),
    iomdkbc_match, iomdkbc_attach, NULL, NULL);

static int
iomdkbc_match(struct device *parent, struct cfdata *cf, void *aux)
{
	struct kbd_attach_args *ka = aux;
	struct opms_attach_args *pa = aux;

	if (strcmp(ka->ka_name, "kbd") == 0 ||
	    strcmp(pa->pa_name, "opms") == 0)
		return 1;
	return 0;
}

static void
iomdkbc_attach(struct device *parent, struct device *self, void *aux)
{
	struct kbd_attach_args *ka = aux;
	struct opms_attach_args *pa = aux;
	struct iomdkbc_softc *sc = (struct iomdkbc_softc *)self;

	sc->sc_pt.t_xt_translation = iomdkbc_xt_translation;
	sc->sc_pt.t_send_devcmd = iomdkbc_send_devcmd;
	sc->sc_pt.t_poll_data1 = iomdkbc_poll_data1;
	sc->sc_pt.t_slot_enable = iomdkbc_slot_enable;
	sc->sc_pt.t_intr_establish = iomdkbc_intr_establish;
	sc->sc_pt.t_controller = &sc->sc_dev;

	printf("\n");

	if (strcmp(ka->ka_name, "kbd") == 0) {
		sc->sc_haveport[PCKBPORT_KBD_SLOT] = 1;
		sc->sc_iot = ka->ka_iot;
		sc->sc_ioh[PCKBPORT_KBD_SLOT] = ka->ka_ioh;
		sc->sc_rxih[PCKBPORT_KBD_SLOT] = intr_claim(ka->ka_rxirq,
		    IPL_TTY, sc->sc_dev.dv_xname, iomdkbc_intr, sc);
		sc->sc_rxirq[PCKBPORT_KBD_SLOT] = ka->ka_rxirq;
		disable_irq(sc->sc_rxirq[PCKBPORT_KBD_SLOT]);
		pckbport_attach_slot(&sc->sc_pt, PCKBPORT_KBD_SLOT);
	}

	if (strcmp(pa->pa_name, "opms") == 0) {
		sc->sc_haveport[PCKBPORT_AUX_SLOT] = 1;
		sc->sc_iot = pa->pa_iot;
		sc->sc_ioh[PCKBPORT_AUX_SLOT] = pa->pa_ioh;
		sc->sc_rxih[PCKBPORT_AUX_SLOT] = intr_claim(pa->pa_irq,
		    IPL_TTY, sc->sc_dev.dv_xname, iomdkbc_intr, sc);
		sc->sc_rxirq[PCKBPORT_AUX_SLOT] = pa->pa_irq;
		disable_irq(sc->sc_rxirq[PCKBPORT_AUX_SLOT]);
		pckbport_attach_slot(&sc->sc_pt, PCKBPORT_AUX_SLOT);
	}
}

static int
iomdkbc_send_devcmd(struct device *self, pckbport_slot_t slot, u_char cmd)
{
	struct iomdkbc_softc *sc = (void *)self;
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_ioh[slot];
	int timeout;

	timeout = 10000;
	while ((bus_space_read_1(iot, ioh, IOMDKBC_CR) &
		   IOMDKBC_TXE) == 0) {
		DELAY(10);
		if (--timeout == 0) return 0;
	}

        bus_space_write_1(iot, ioh, IOMDKBC_DR, cmd);
	return 1;
}

static int
iomdkbc_poll_data1(struct device *self, pckbport_slot_t slot)
{
	struct iomdkbc_softc *sc = (void *)self;
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_ioh[slot];
	int timeout;

	timeout = 10000;
	while ((bus_space_read_1(iot, ioh, IOMDKBC_CR) &
		   IOMDKBC_RXF) == 0) {
		DELAY(10);
		if (--timeout == 0) return -1;
	}

        return bus_space_read_1(iot, ioh, IOMDKBC_DR);
}

/*
 * switch scancode translation on / off
 * return nonzero on success
 */
static int
iomdkbc_xt_translation(struct device *self, pckbport_slot_t slot, int on)
{

	if (on)
		return 0; /* Can't do XT translation */
	else
		return 1;
}

static void
iomdkbc_slot_enable(struct device *self, pckbport_slot_t slot, int on)
{
	struct iomdkbc_softc *sc = (void *)self;
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_ioh[slot];

	bus_space_write_1(iot, ioh, IOMDKBC_CR, on ? IOMDKBC_ENA : 0);
}


static void
iomdkbc_intr_establish(struct device *self, pckbport_slot_t slot)
{
	struct iomdkbc_softc *sc = (void *)self;

	enable_irq(sc->sc_rxirq[slot]);
}

static int
iomdkbc_intr(void *self)
{
	struct iomdkbc_softc *sc = (void *)self;
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh;
	unsigned i;
	int stat, ret;

	ret = 0;
	for (i = 0; i < PCKBPORT_NSLOTS; i++)
		if (sc->sc_haveport[i]) {
			ioh = sc->sc_ioh[i];
			stat = bus_space_read_1(iot, ioh, IOMDKBC_CR);
			if ((stat & IOMDKBC_RXF) == 0)
				continue;
			pckbportintr(&sc->sc_pt, i,
			    bus_space_read_1(iot, ioh, IOMDKBC_DR));
			ret = 1;
		}

	return ret;
}

/* Dummy softc used before autoconfig */
static struct iomdkbc_softc console_sc;

int
iomdkbc_cnattach(bus_space_tag_t iot, bus_addr_t addr, int slot)
{
	struct iomdkbc_softc *sc = &console_sc;

	sc->sc_iot = iot;
	bus_space_map(iot, addr, 2, 0, &sc->sc_ioh[slot]);
	sc->sc_pt.t_xt_translation = iomdkbc_xt_translation;
	sc->sc_pt.t_send_devcmd = iomdkbc_send_devcmd;
	sc->sc_pt.t_poll_data1 = iomdkbc_poll_data1;
	sc->sc_pt.t_slot_enable = iomdkbc_slot_enable;
	sc->sc_pt.t_intr_establish = iomdkbc_intr_establish;
	sc->sc_pt.t_controller = &sc->sc_dev;
	strcpy(sc->sc_dev.dv_xname, "iomdkbc");
	sc->sc_haveport[slot] = 1;
	iomdkbc_slot_enable(&sc->sc_dev, slot, 1);
	return pckbport_cnattach(&sc->sc_pt, slot);
}
